%{
#include "minako.h"
#include "codegen.h"
#include <stdlib.h>
#include <assert.h>
#include <string.h>
	
#define YYERROR_VERBOSE 1
#define YYDEBUG 1
	
	/* parserinterne Variablen */
	static symtab* table;
	static codegen* gen;
	
/* ******************************************************************* Intern */
	
	/* Zusatzfunktionen, die das Handling von Variablen und Funktionen
	 * erleichtern */
	
	static minakoVariable* getMinakoVariable(symtab* table, const char* id);
	static minakoFunction* getMinakoFunction(symtab* table, const char* id);
	static const char* getType(long type);
	static int combineTypes(int a, int b, int commutative);
	
%}

%union {
	double floatValue;
	int intValue;
	char* string;
	stack* param;
}

/* benutzte Terminalsymbole */
%token AND
%token OR
%token EQ
%token NEQ
%token LEQ
%token GEQ
%token LSS
%token GRT
%token KW_BOOLEAN
%token KW_DO
%token KW_ELSE
%token KW_FLOAT
%token KW_FOR
%token KW_IF
%token KW_INT
%token KW_PRINTF
%token KW_RETURN
%token KW_VOID
%token KW_WHILE
%token <intValue> CONST_INT
%token <floatValue> CONST_FLOAT
%token <intValue> CONST_BOOLEAN
%token <string> CONST_STRING
%token <string> ID

/* Definition der Operatorassoziativitäten und -präzedenzen */
%nonassoc LOWER_THAN_ELSE
%nonassoc KW_ELSE
%left AND
%left OR
%left EQ NEQ GEQ LEQ GRT LSS
%left '+' '-'
%left '*' '/'
%nonassoc UMINUS

/* Definition der Typen der Nichtterminalsymbole */
%type <param> lop_actualparameter
%type <param> lop_formalparameter
%type <param> opt_formalparameter
%type <intValue> functioncall
%type <intValue> type
%type <intValue> assignment
%type <intValue> expr

%%

program:
	/* leer */ {
		/* testen, ob die Funktion "void main()" deklariert worden ist */	
		minakoFunction* func = getMinakoFunction(table, "main");
		
		/* neu setzen der Zeilennummer für eine genauere Fehlerausgabe */
		yylineno = func->line;
		
		if (func->type != 'void')
			yyerror("Rückgabetyp der Funktion \"main\" muss void sein.");
		
		if (func->argCount)
			yyerror("\"main\" darf keine Parameter haben.");
	}
	| declassignment ';' program
	| functiondefinition program
	;

functiondefinition:
	type ID '(' opt_formalparameter ')' {
		/* Funktion deklarieren */
		symtabSymbol* sym = newMinakoFunction($1, $2);
		minakoFunction* func = (minakoFunction*) sym->data;
		int i = 0;
		
		minakoFunctionDeclare(table, sym, $4);
		deleteSymtabSymbol(sym);
		
		/* neuer Scope für die lokalen Variablen */
		symtabPush(table);
		
		/* teile der Codegenerierungskomponente das betreten der Funktion mit */
		codegenEnterFunction(gen, $2, func);
		
		/* speichere den Namen der aktuellen Funktion in der Symboltabelle */
		sym = newSymtabSymbol('meta', "$func");
		sym->data = strdup($2);
		symtabDeclare(table, sym);
		deleteSymtabSymbol(sym);
		
		/* deklarieren der Parametervariablen */
		while (i < func->argCount)
		{
			minakoVariableDeclare(table, func->param[i]);
			++i;
		}
	} '{' statementlist '}' {
		codegenLeaveFunction(gen);
		symtabPop(table);
	}
	;

opt_formalparameter:
	/* leer */                    { $$ = newStack(); }
	| type ID lop_formalparameter {
		stackPush($3, newMinakoVariable($1, $2));
		
		if ($1 == 'void')
			yyerror("Ein Funktionsparameter kann nicht vom Typ void sein.");
		
		$$ = $3;
	}
	;

lop_formalparameter:
	/* leer */                         { $$ = newStack(); }
	| ',' type ID lop_formalparameter  {
		stackPush($4, newMinakoVariable($2, $3));
		
		if ($2 == 'void')
			yyerror("Ein Funktionsparameter kann nicht vom Typ void sein.");
		
		$$ = $4;
	}
	;

functioncall:
	ID '(' ')' {
		minakoFunction* func = getMinakoFunction(table, $1);
		
		if (func->argCount)
			yyerror("\"%s\" muss mit Parametern gerufen werden.", $1);
		
		$$ = func->type;
	}
	| ID '(' assignment lop_actualparameter ')' {
		minakoFunction* func = getMinakoFunction(table, $1);
		int i = 0;
		int type;
		
		/* den ersten Parameter auf den Stack pushen */
		stackPush($4, (void*)$3);
		
		/* Funktionsargumente auf Typverträglichkeit und Anzahl prüfen */
		do
		{
			minakoVariable* var;
			
			if (i == func->argCount)
				yyerror("beim Aufruf von \"%s\": Zu viele Parameter übergeben.", $1);
			
			var = (minakoVariable*) func->param[i]->data;
			type = (int) stackPop($4);
			
			if (!combineTypes(var->type, type, 0))
				yyerror("beim Aufruf von \"%s\": Ein %s kann keinem %s zugewiesen "
				        "werden.", $1, getType(type), getType(var->type));
			
			++i;
		}
		while (!stackIsEmpty($4));
		
		deleteStack($4);
		
		if (i < func->argCount)
			yyerror("beim Aufruf von \"%s\": %i statt %i Parameter übergeben.", $1, i,
			        func->argCount);
		
		$$ = func->type;
	}
	;

lop_actualparameter:
	/* leer */                           { $$ = newStack(); }
	| ',' assignment lop_actualparameter {
		stackPush($3, (void*)$2);
		$$ = $3;
	}
	;

statementlist:
	/* leer */
	| statementlist block
	;

block: 
	'{' {
		symtabPush(table);
	} statementlist '}' {
		symtabPop(table);
	}
	| statement
	;

block2: 
	'{' {
		symtabPush(table); 
	} statementlist '}' {
		symtabPop(table);
	}
	| statement
	;

statement: 
	  ifstatement 
	| forstatement 
	| whilestatement 
	| returnstatement ';' 
	| dowhilestatement ';' 
	| printf ';' 
	| declassignment ';' 
	| statassignment ';' 
	| functioncall ';'
	;

ifstatement: 
	ifcondition block2 %prec LOWER_THAN_ELSE
	| ifcondition block2 KW_ELSE block2
	;
	
ifcondition:
	KW_IF '(' assignment ')' {
		if ($3 != 'bool')
			yyerror("Die Bedingung muss vom Typ Boolean sein.");
	}
	;

forstatement: 
	KW_FOR '(' declassignment ';' expr {
		if ($5 != 'bool')
			yyerror("Die Bedingung in der for-Schleife muss vom Typ Boolean sein.");
	} ';' statassignment ')' block2
	| KW_FOR '(' statassignment ';' expr {
		if ($5 != 'bool')
			yyerror("Die Bedingung in der for-Schleife muss vom Typ Boolean sein.");
	} ';' statassignment ')' block2
	;

dowhilestatement: 
	KW_DO block2 KW_WHILE '(' assignment ')' {
		if ($5 != 'bool')
			yyerror("Die Bedingung muss vom Typ Boolean sein.");
	}
	;

whilestatement: 
	KW_WHILE '(' assignment ')' {
		if ($3 != 'bool')
			yyerror("Die Bedingung muss vom Typ Boolean sein."); 
	} block2
	;

returnstatement: 
	KW_RETURN {
		/* aktuell geparste Funktion ermitteln */
		const symtabSymbol* sym = symtabGet(table, "$func");
		minakoFunction* func = getMinakoFunction(table,
		                                         (const char*) sym->data);
		
		if (func->type != 'void')
			yyerror("Die Funktion \"%s\" muss %s zurückgeben.", (char*) sym->data,
			        getType(func->type));
	}
	| KW_RETURN assignment {
		/* aktuell geparste Funktion ermitteln */
		const symtabSymbol* sym = symtabGet(table, "$func");
		minakoFunction* func = getMinakoFunction(table,
		                                         (const char*) sym->data);
		
		/* Test auf Typverträglichkeit */
		if (!combineTypes(func->type, $2, 0))
			yyerror("Die Funktion \"%s\" sollte %s statt %s zurückgeben.",
			        (char*) sym->data, getType(func->type), getType($2));
	}
	;

printf: 
	KW_PRINTF '(' assignment ')' 
	| KW_PRINTF '(' CONST_STRING ')'
	;

declassignment: 
	type ID {
		symtabSymbol* sym = newMinakoVariable($1, $2);
		
		if ($1 == 'void')
			yyerror("Eine Variable kann nicht vom Typ void sein.");
		
		minakoVariableDeclare(table, sym);
		deleteSymtabSymbol(sym);
	}
	| type ID '=' assignment {
		symtabSymbol* sym = newMinakoVariable($1, $2);
		
		if (!combineTypes($1, $4, 0))
			yyerror("Ein %s kann keinem %s zugewiesen werden.", getType($4),
			        getType($1));
		
		minakoVariableDeclare(table, sym);
		deleteSymtabSymbol(sym);
	}
	;

type:
	KW_BOOLEAN { $$ = 'bool'; } 
	| KW_FLOAT { $$ = 'real'; }
	| KW_INT   { $$ = 'int'; }
	| KW_VOID  { $$ = 'void'; }
	;

statassignment: 
	ID '=' assignment {
		minakoVariable* var = getMinakoVariable(table, $1);
		
		if (!combineTypes(var->type, $3, 0))
			yyerror("Ein %s kann keinem %s zugewiesen werden.", getType($3),
			        getType(var->type));
	}
	;

assignment:
	ID '=' assignment {
		minakoVariable* var = getMinakoVariable(table, $1);
		
		if (!($$ = combineTypes(var->type, $3, 0)))
			yyerror("Ein %s kann keinem %s zugewiesen werden.", getType($3),
			        getType(var->type));
	}
	| expr { $$ = $1; }
	;

expr:
	expr '+' expr {
		if (!($$ = combineTypes($1, $3, 1)))
			yyerror("Ein %s kann nicht zu einem %s addiert werden.",
			        getType($1), getType($3));
	}
	| expr '-' expr {
		if (!($$ = combineTypes($1, $3, 1)))
			yyerror("Ein %s kann von zu einem %s subtrahiert werden.",
			        getType($1), getType($3));
	}
	| expr '*' expr {
		if (!($$ = combineTypes($1, $3, 1)))
			yyerror("Ein %s kann nicht mit einem %s multipliziert werden.",
			        getType($1), getType($3));
	}
	| expr '/' expr {
		if (!($$ = combineTypes($1, $3, 1)))
			yyerror("Ein %s kann nicht durch ein %s dividiert werden.",
			        getType($1), getType($3));
	}
	| '-' expr %prec UMINUS {
		if ($2 != 'int' && $2 != 'real')
			yyerror("Ein %s kann nicht negiert werden.", getType($2));
		
		$$ = $2;
	}
	| expr AND expr {
		if ($1 != 'bool' || $3 != 'bool')
			yyerror("Nur boolesche Ausdrücke können mit && verknüpft werden.");
		
		$$ = 'bool';
	}
	| expr OR expr {
		if ($1 != 'bool' || $3 != 'bool')
			yyerror("Nur boolesche Ausdrücke können mit || verknüpft werden.");
		
		$$ = 'bool';
	}
	| expr EQ expr {
		if (!combineTypes($1, $3, 1))
			yyerror("Ein %s kann nicht auf Gleichheit mit einem %s getestet werden",
			        getType($1), getType($3));
		
		$$ = 'bool';
	}
	| expr NEQ expr {
		if (!combineTypes($1, $3, 1))
			yyerror("Ein %s kann nicht auf Ungleichheit mit einem %s getestet werden",
			        getType($1), getType($3));
		
		$$ = 'bool';
	}
	| expr LEQ expr {
		if ($1 == 'bool' || !combineTypes($1, $3, 1))
			yyerror("Der Größenvergleich zwischen %s und %s ist undefiniert.",
			        getType($1), getType($3));
		
		$$ = 'bool';
	}
	| expr GEQ expr {
		if ($1 == 'bool' || !combineTypes($1, $3, 1))
			yyerror("Der Größenvergleich zwischen %s und %s ist undefiniert.",
			        getType($1), getType($3));
		
		$$ = 'bool';
	}
	| expr LSS expr {
		if ($1 == 'bool' || !combineTypes($1, $3, 1))
			yyerror("Der Größenvergleich zwischen %s und %s ist undefiniert.",
			        getType($1), getType($3));
		
		$$ = 'bool';
	}
	| expr GRT expr {
		if ($1 == 'bool' || !combineTypes($1, $3, 1))
			yyerror("Der Größenvergleich zwischen %s und %s ist undefiniert.",
			        getType($1), getType($3));
		
		$$ = 'bool';
	}
	| CONST_INT          { $$ = 'int'; }
	| CONST_FLOAT        { $$ = 'real'; }
	| CONST_BOOLEAN      { $$ = 'bool'; }
	| functioncall       { $$ = $1; }
	| ID                 { $$ = getMinakoVariable(table, $1)->type; }
	| '(' assignment ')' { $$ = $2; }
	;

%%

int main(int argc, char* argv[])
{
	table = newSymtab(0);
	gen = newCodegen("MinaOut");
	
	if (argc < 2)
	{
		yyin = stdin;
	}
	else
	{
		yyin = fopen(argv[1], "r");
		if (!yyin)
			printf("Fehler: Datei %s nicht gefunden", argv[0]);
	}
	
	yyparse();
	
	deleteSymtab(table);
	deleteCodegen(gen);
	
	return 0;
}


#define NEW(_struct) (_struct*)malloc(sizeof(_struct))
	
/* ********************************************************* Minako Variablen */
	
	symtabSymbol* newMinakoVariable(long type, const char* id)
	{
		symtabSymbol* sym = newSymtabSymbol('var', id);
		minakoVariable* var = NEW(minakoVariable);
		
		var->type = type;
		var->line = yylineno;
		
		sym->data = var;
		sym->destructor = (fDestructor) deleteMinakoVariable;
		
		return sym;
	}
	
	void deleteMinakoVariable(minakoVariable* var)
	{
		codegenRemoveVariable(gen, var);
		free(var);
	}
	
	void minakoVariableDeclare(symtab* table, const symtabSymbol* symb)
	{
		assert(symb && symb->class == 'var');
		
		/* teste, ob die Variable schon existiert */
		if (!symtabDeclare(table, symb))
		{
			/* suche das Originalsymbol für eine schönere Fehlermeldung */
			symb = symtabGet(table, symb->id);
			
			switch (symb->class)
			{
			case 'var':
				yyerror("Die Variable \"%s\" wurde bereits in Zeile %i deklariert.",
				        symb->id, ((minakoVariable*)symb->data)->line);
				
			case 'func':
				yyerror("Es existiert bereits eine Funktion in Zeile %i mit dem Namen "
				        "\"%s\".", ((minakoFunction*)symb->data)->line, symb->id);
				
			default:
				yyerror("Der Bezeichner \"%s\" ist bereits vergeben.", symb->id);
			}
		}
		
		/* registriere die Variable im Codegenerator */
		codegenAddVariable(gen, symb->id, (minakoVariable*) symb->data);
	}
	
/* ******************************************************** Minako Funktionen */
	
	symtabSymbol* newMinakoFunction(long type, const char* id)
	{
		symtabSymbol* symb = newSymtabSymbol('func', id);
		minakoFunction* func = NEW(minakoFunction);
		
		func->type = type;
		func->line = yylineno;
		
		symb->data = func;
		symb->destructor = (fDestructor) deleteMinakoFunction;
		
		return symb;
	}
	
	void deleteMinakoFunction(minakoFunction* func)
	{
		int i = func->argCount;
		
		while (i--)
		{
			deleteMinakoVariable((minakoVariable*) func->param[i]->data);
			deleteSymtabSymbol(func->param[i]);
		}
		
		free(func->param);
		free(func);
	}
	
	void minakoFunctionDeclare(symtab* table, const symtabSymbol* symb,
	                           stack* parameter)
	{
		minakoFunction* func;
		stack* p;
		int count;
		
		assert(symb && symb->class == 'func' && parameter);
		
		func = (minakoFunction*) symb->data;
		
		/* prüfe, ob die Funktion schon in der Symboltabelle existiert */
		if (!symtabDeclare(table, symb))
		{
			/* für eine schönere Fehlerausgabe, suche das Originalsymbol */
			symb = symtabGet(table, symb->id);
			
			switch (symb->class)
			{
			case 'var':
				yyerror("Es existiert bereits eine Variable in Zeile %i mit dem Namen "
				        "\"%s\".", ((minakoVariable*)symb->data)->line, symb->id);
				
			case 'func':
				yyerror("Die Funktion \"%s\" wurde bereits in Zeile %i deklariert.",
				        symb->id,	((minakoFunction*)symb->data)->line);
				
			default:
				yyerror("Der Bezeichner \"%s\" ist bereits vergeben.", symb->id);
			}
		}
		
		/* zähle die Elemente auf dem Stack */
		p = newStack();
		count = 0;
		
		while (!stackIsEmpty(parameter))
		{
			stackPush(p, stackPop(parameter));
			++count;
		}
		
		deleteStack(parameter);
		
		/* nachdem die Anzahl bekannt ist, ist ein statisches Array günstiger */
		func->param = (symtabSymbol**) malloc(sizeof(symtabSymbol*) * count);
		func->argCount = count;
		
		while (count--)
		{
			symtabSymbol* s = (symtabSymbol*)stackPop(p);
			func->param[count] = s;
			
			/* Destruktor auf 0 setzen führt dazu, dass die Parameter nicht von der
			 * Symboltabelle freigegeben werden - wir kümmern uns selbst darum */
			s->destructor = 0;
		}
		
		deleteStack(p);
	}
	
#undef NEW
	
	static minakoVariable* getMinakoVariable(symtab* table, const char* id)
	{
		const symtabSymbol* s = symtabGet(table, id);
		
		if (!s)
			yyerror("Variable \"%s\" ist nicht deklariert worden.", id);
		
		if (s->class != 'var')
			yyerror("\"%s\" kann nicht wie eine Variable benutzt werden.", id);
		
		return (minakoVariable*)s->data;
	}
	
	static minakoFunction* getMinakoFunction(symtab* table, const char* id)
	{
		const symtabSymbol* s = symtabGet(table, id);
		
		if (!s)
			yyerror("Funktion \"%s\" ist nicht deklariert worden.", id);
		
		if (s->class != 'func')
			yyerror("\"%s\" kann nicht wie eine Funktion benutzt werden.", id);
		
		return (minakoFunction*)s->data;
	}
	
	static const char* getType(long type)
	{
		switch (type)
		{
		case 'int':
			return "int";
			
		case 'real':
			return "float";
			
		case 'bool':
			return "bool";
			
		case 'void':
			return "void";
			
		default:
			return "ILLEGAL";
		}
	}
	
	static int combineTypes(int a, int b, int commutative)
	{
		switch (a)
		{
		case 'int':
			switch (b)
			{
			case 'int':
				return 'int';
				
			case 'real':
				if (commutative)
					return 'real';
			
			default:
				return 0;
			}
			
		case 'real':
			switch (b)
			{
			case 'int':
			case 'real':
				return 'real';
				
			default:
				return 0;
			}
			
		case 'bool':
			switch (b)
			{
			case 'bool':
				return 'bool';
				
			default:
				return 0;
			}
			
		default:
			return 0;
		}
	}
