#include "symtab.h"
#include "stack.h"
#include "hashmap.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#define NEW(_STRUCT) (_STRUCT*)malloc(sizeof(_STRUCT))

/* *************************************************************** Strukturen */

struct sSymtabSymbol
{
	symtabSymbol symb; /**< Symbol mit Daten */
	stack* layerPtr; /**< Zeiger auf die definierende Schicht */
};

struct sSymtab
{
	hashmap* symbolMap; /**< Symbolmap */
	stack* layerStack; /**< Schichtenstack */
};

struct sIdStackTuple
{
	struct sSymtabSymbol *this, *last;
};

/* ******************************************************* private Funktionen */

/**@brief Entfernt eine Deklarationsschicht aus der Symboltabelle.
 */
static void revert(stack* history, hashmap* symbolMap);

/* *************************************************** exportierte Funktionen */

symtab* newSymtab(unsigned int hint)
{
	symtab* table = NEW(symtab);
	
	table->symbolMap = newHashmap(hint);
	table->layerStack = newStack();
	
	/* lege die globale Schicht an */
	stackPush(table->layerStack, newStack());
	
	return table;
}

void deleteSymtab(symtab* table)
{
	assert(table);
	
	/* entferne alle Schichten einzeln um Memory-Leaks zu verhindern */
	while (!stackIsEmpty(table->layerStack))
	{
		revert((stack*) stackPop(table->layerStack), table->symbolMap);
	}
	
	/* und nun den Rest */
	deleteHashmap(table->symbolMap);
	deleteStack(table->layerStack);
	free(table);
}

void symtabPush(symtab* table)
{
	assert(table);
	stackPush(table->layerStack, newStack());
}

void symtabPop(symtab* table)
{
	stack* identifierStack;
	assert(table);
	identifierStack = (stack*) stackPop(table->layerStack);
	
	/* die allererste Schicht darf nicht entfernt werden */
	assert(!stackIsEmpty(table->layerStack));
	
	/* nun setze die Variablen auf ihre alten Werte zurück */
	revert(identifierStack, table->symbolMap);
}

symtabSymbol* newSymtabSymbol(int class, const char* id)
{
	symtabSymbol* symb = NEW(symtabSymbol);
	
	symb->class = class;
	symb->destructor = free;
	symb->data = 0;
	symb->id = strdup(id);
	
	return symb;
}

void deleteSymtabSymbol(symtabSymbol* symb)
{
	free((char*)symb->id);
	free(symb);
}

int symtabDeclare(symtab* table, const symtabSymbol* symb)
{
	struct sSymtabSymbol* s;
	struct sIdStackTuple* t;
	stack* history;
	int result;
	
	assert(table && symb);
	
	s = (struct sSymtabSymbol*) hashmapGet(table->symbolMap, symb->id);
	history = (stack*) stackTop(table->layerStack);
	
	/* teste, ob der Bezeichner schon existiert */
	if (s)
	{
		/* teste, ob er (nicht) in der aktuellen Schicht deklariert wurde */
		if (s->layerPtr != history)
		{
			/* speichere das alte Symbol */
			struct sSymtabSymbol* copy = NEW(struct sSymtabSymbol);
			t = NEW(struct sIdStackTuple);
			
			/* überschreibe das aktuelle */
			memcpy(copy, s, sizeof(struct sSymtabSymbol));
			
			/* setze das Vorgängerelement */
			t->last = copy;
			result = 2;
		}
		else
			return 0;
	}
	else
	{
		/* lege ein neues Symbol an */
		s = NEW(struct sSymtabSymbol);
		t = NEW(struct sIdStackTuple);
		
		/* kopiere den Schlüssel */
		s->symb.id = strdup(symb->id);
		
		hashmapSet(table->symbolMap, s, symb->id);
		
		/* setze das Vorgängerelement (kein Vorgänger) */
		t->last = 0;
		result = 1;
	}
	
	/* setze das aktuelle Element und lege es auf den Stack */
	t->this = s;
	stackPush(history, t);
	
	/* setze die neuen Symboleigenschaften */
	s->layerPtr = history;
	s->symb.class = symb->class;
	s->symb.data = symb->data;
	s->symb.destructor = symb->destructor;
	
	return result;
}

const symtabSymbol* symtabGet(const symtab* table, const char* id)
{
	struct sSymtabSymbol* s;
	
	assert(table);
	
	s = hashmapGet(table->symbolMap, id);
	
	if (s)
		return &s->symb;
	
	return 0;
}

void symtabPrint(const symtab* table, fSymtabPrinter printer)
{
	assert(table);
	
	hashmapPrint(table->symbolMap, (fHashmapPrinter)printer);
}

/* ********************************************************** Hilfsfunktionen */

static void revert(stack* history, hashmap* symbolMap)
{
	/* iteriere über alle veränderten Symbole dieser Schicht */
	while (!stackIsEmpty(history))
	{
		struct sIdStackTuple* datum = (struct sIdStackTuple*) stackPop(history);
		
		/* gib den eventuell benutzten Speicher frei */
		if (datum->this->symb.destructor && datum->this->symb.data)
			datum->this->symb.destructor(datum->this->symb.data);
		
		/* teste, ob es eine ältere Definition dieses Typs gegeben hat */
		if (datum->last)
		{
			/* setze die Definition zurück */
			memcpy(datum->this, datum->last, sizeof(struct sSymtabSymbol));
		}
		else
		{
			/* entferne die Definition dieses Symboles aus der Symboltabelle */
			hashmapRemove(symbolMap, datum->this->symb.id);
			
			datum->last = datum->this;
			
			/* gib den Speicher für den String frei */
			free((char*)datum->this->symb.id);
		}
		
		/* aufräumen des betrachteten Symbols */
		free(datum->last);
		free(datum);
	}
	
	deleteStack(history);
}
